# -*- coding: utf-8 -*-
#
"""
Deal with configuration and data files.
Parse configuration file and set defaults.
"""

#    Copyright (c) 2009 Efrim <efrim@azylum.org> {{{
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, version 3 of the License only.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  }}}

# DOC {{{
#    http://mail.python.org/pipermail/python-list/2004-February/248674.html
# }}}

# IMPORTS {{{
from __future__ import with_statement
import ConfigParser
import pickle
import sys

from ConfigParser import (MissingSectionHeaderError, ParsingError)
from os import (makedirs, environ, stat, chmod, rename)
from os.path import (join, isdir, isfile, expanduser)
from stat import (S_IMODE, ST_MODE, S_IRWXO, S_IRWXG)
#}}}

#DEFAULTS{{{
CONF_FILE = 'mpd_sima.cfg'
USR_DB = 'user_db.cfg'

# Mandatory user options
M_SECTIONS = ['MPD']
# Optional user options
O_SECTIONS = ['sima', 'log']

DEFAULT_CONF = {
        'MPD': {
            'host': "localhost",
            'port': "6600",
            'password': "false"},
        'sima': {
            'top_tracks': "false",
            'history': "false",
            'cache': "false",
            'user_db': "false",
            'queue_length': "2",
            'track_to_add': "1",
            'main_loop_time': "4"},
        'log': {
            'verbosity': "info"}}

#}}}

class CacheError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


#CONFIG FONCTIONS{{{

def load_cache(cache_file):#{{{
    """Load cache from file."""
    similar_cache = {}
    cache_stamp = None
    if isfile(cache_file):
        with open(cache_file, 'rb') as cache_fd:
            similar_cache = pickle.load(cache_fd)
            cache_stamp = pickle.load(cache_fd)

    return similar_cache, cache_stamp#}}}

def load_usr_cache(cache_file):
    """Add User Cache to regular cache"""
    config = ConfigParser.SafeConfigParser()
    try:
        config.read(cache_file)
    except (MissingSectionHeaderError, ParsingError), err_conf:
        raise CacheError(err_conf)

    usr_db = {}
    # Convert config to dict object to match CACHE type
    for artist in config.sections():
        usr_db[artist] = [ art[0] for art in config.items(artist) ]

    return usr_db
#}}}


class ConfMan():#CONFIG MANAGER CLASS{{{
    """
    Configuration manager.
    get_config is run first and retrieve configuration from defaults if not set
    in conf files.
    """

    def __init__(self, logger):#{{{
        self.conf_file = None
        self.history_file = None
        self.similar_cache = None
        self.user_cache = None
        self.config = None
        self.defaults = DEFAULT_CONF
        self.sima_logger = logger
        self.get_config()
        self.control_mod()
        self.is_mpdenvvar_changed()#}}}

    def get_mpd_environ(self):#{{{
        """
        Retrieve MPD env. var.
        """
        mpd_host_and_pass = environ.get('MPD_HOST')
        if mpd_host_and_pass:
            mpd_host_and_pass = mpd_host_and_pass.split('@')

        mpd_port = environ.get('MPD_PORT')
        return (mpd_host_and_pass, mpd_port)#}}}

    def use_mpd_environ(self):#{{{unicodeReady
        """
        Use MPD_HOST/MPD_PORT if set as environment variables.
        """
        mpd_host_pass, mpd_port = self.get_mpd_environ()
        if mpd_host_pass:
            self.sima_logger.info(u'MPD_HOST set to "%s", setting up conf file.' %
                                  mpd_host_pass[0])
            self.defaults['MPD']['host'] = mpd_host_pass[0]
            if len(mpd_host_pass) == 2 and mpd_host_pass[1]:
                self.sima_logger.info(u'MPD_HOST contains password, setting up conf file.')
                self.defaults['MPD']['password'] = mpd_host_pass[1]

        if mpd_port:
            self.sima_logger.info(u'MPD_PORT set to "%i", setting up conf file.'
                                  % int(environ.get('MPD_PORT')))
            self.defaults['MPD']['port'] = environ.get('MPD_PORT') #}}}

    def dict2conf(self, conf_dict):#{{{
        """format conf file lines to be written."""
        a_str = "[%s]\n" % conf_dict
        for key, val in self.defaults[conf_dict].iteritems():
            a_str = a_str + "%s = %s\n" % (key, val)
        a_str = a_str + "\n"
        return a_str#}}}

    def control_mod(self):#{{{unicodeReady
        """
        Controls conf file permissions.
        """
        mode = S_IMODE(stat(self.conf_file)[ST_MODE])
        self.sima_logger.debug(u'file permision is: %o' % mode)
        if mode & S_IRWXO or mode & S_IRWXG:
            self.sima_logger.warning(u'File is readable by "other" and/or'+
                                     u' "group" (actual permission %o octal).' %
                                     mode)
            self.sima_logger.warning(u'Consider setting permissions'+
                                     u' to 600 octal.')
        #}}}

    def config_write_default(self):#{{{unicodeReady
        """Write a default configuration file."""
        ##DEBUG MESS
        self.sima_logger.info(\
                u'No configuration file available, creating one…')
        self.sima_logger.info(\
                u'Configuration file: "%s"' % self.conf_file)
        #self.sima_logger.warning(\
        #        'You may need to tune your MPD host settings…')
        self.use_mpd_environ()
        with open(self.conf_file, 'wu') as fd_conf:
            for section in M_SECTIONS:
                # TODO add a try, what if not writable destination
                fd_conf.write(self.dict2conf(section))
        chmod(self.conf_file, 0600)
        #}}}

    def is_mpdenvvar_changed(self):#{{{unicodeReady
        """
        Check whether or not MPD environement variables differ from conf file.
        """
        mpd_host_pass, mpd_port = self.get_mpd_environ()
        if mpd_host_pass and \
           mpd_host_pass[0].strip() != self.config.get('MPD', 'host'):
            self.sima_logger.warning(u'MPD_HOST env. var. is set to "%s"' %
                                     mpd_host_pass[0])
            self.sima_logger.warning(u'Configuration file refers to "%s" (using it).' %
                                     self.config.get('MPD', 'host'))
        # Then test if password is set. If a Password is set then host must
        # be specified (man 1 mpc).
        # MPD_HOST = host@password
        if mpd_host_pass and \
           len(mpd_host_pass) == 2 and \
           mpd_host_pass[1].strip() != self.config.get('MPD', 'password'):
            self.sima_logger.warning(u'Password in MPD_HOST env. var. differs'+
                                     u' from password in configuration file (use configuration).')
        if mpd_port and \
           mpd_port.strip() != self.config.get('MPD', 'port'):
            self.sima_logger.warning(u'MPD_PORT env. var. is set to "%i"' %
                                     int(mpd_port))
            self.sima_logger.warning(u'Configuration file refers to "%i" (using it).' %
                                     self.config.getint('MPD', 'port'))#}}}

    def control_conf(self):#{{{unicodeReady
        """Get through options/values and set defaults if not in conf file."""
        # Optional sections for user conf
        # Setting default if not specified
        for section in O_SECTIONS:
            if section not in self.config.sections():
                self.sima_logger.debug(u'[%s] NOT in conf file' % section)
                self.config.add_section(section)
                for option in DEFAULT_CONF[section]:
                    self.config.set(section,
                            option,
                            DEFAULT_CONF[section][option])
                    self.sima_logger.debug(
                            u'Setting option with default value: %s = %s' %
                            (option, DEFAULT_CONF[section][option]))
            elif section in self.config.sections():
                self.sima_logger.debug(u'[%s] present in conf file' % section)
                for option in DEFAULT_CONF[section]:
                    if self.config.has_option(section, option):
                        pass
                    else:
                        self.sima_logger.debug(
                                u'Option "%s" missing in section "%s"' %
                                (option, section))
                        self.sima_logger.debug(u'=> setting default "%s" (may not suit you…)' %
                                               DEFAULT_CONF[section][option])
                        self.config.set(section, option,
                                        DEFAULT_CONF[section][option])
        # Mandatory options
        for section in M_SECTIONS:
            if section not in self.config.sections():
                self.sima_logger.error(u'[%s] NOT in conf file' % section)
                self.config.add_section(section)
                for option in DEFAULT_CONF[section]:
                    self.config.set(section, option, \
                                    DEFAULT_CONF[section][option])
                    self.sima_logger.warning(
                            u'Trying with default values for section "%s"' %
                            section)
                    self.sima_logger.warning(u'Setting to "%s = %s"' %
                            (option, DEFAULT_CONF[section][option]))
            elif section in self.config.sections():
                self.sima_logger.debug(u'[%s] present in conf file' % section)
                for option in DEFAULT_CONF[section]:
                    if self.config.has_option(section, option):
                        pass
                    else:
                        self.sima_logger.warning(
                            u'Option "%s" missing in section "%s"' %
                            (option, section))
                        self.sima_logger.warning(u'setting default "%s" (may not suit you…)' %
                                                 DEFAULT_CONF[section][option])
                        self.config.set(section, option,
                                        DEFAULT_CONF[section][option])

        return self.config#}}}

    def get_config(self):#{{{
        """
        Use XDG directory standard if exists
        else use "HOME/(.config|.local/share)/mpd_sima/"
        http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html
        """

        homedir = expanduser('~')

        if environ.get('XDG_DATA_HOME'):
            data_dir = join(environ.get('XDG_DATA_HOME'), 'mpd_sima')
        else:
            data_dir = join(homedir, '.local', 'share', 'mpd_sima')

        if not isdir(data_dir):
            makedirs(data_dir)
            chmod(data_dir, 0700)

        if environ.get('XDG_CONFIG_HOME'):
            conf_dir = join(environ.get('XDG_CONFIG_HOME'), 'mpd_sima')
        else:
            conf_dir = join(homedir, '.config', 'mpd_sima')

        if not isdir(conf_dir):
            makedirs(conf_dir)
            chmod(conf_dir, 0700)

        self.history_file = join(data_dir, "history.pkl")
        #V0.2.0=>0.3.0
        old_hist = join(data_dir, "add_history.pkl")
        if isfile(old_hist):
            rename(old_hist, self.history_file)
            self.sima_logger.info(u'Old history "add_history.pkl" renamed "history.pkl"')
        self.similar_cache = join(data_dir, "similar_cache.pkl")

        self.conf_file = join(conf_dir, CONF_FILE)
        self.user_cache = join(conf_dir, USR_DB)

        self.sima_logger.info(u'Loading configuration from:  %s' % self.conf_file)
        # Load conf file
        if not isfile(self.conf_file):
            self.config_write_default()

        config = ConfigParser.SafeConfigParser()

        try:
            config.read(self.conf_file)
        except (MissingSectionHeaderError, ParsingError), err_conf:
            self.sima_logger.error(u'error loading %s:\n%s' %
                                   (self.conf_file, err_conf))
            sys.exit(1)

        self.config = config
        #}}}

#}}}

if __name__ == "__main__":#{{{
    pass

#}}}

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab
